from ipycanvas import Canvas

class IpycanvasDrawer:
    def __init__(self, *, width=None, height=None, save_image=None, **kwargs):
        self.canvas = Canvas(width=width, height=height, sync_image_data=True)

        self.canvas.font = "18px Arial"
        self.canvas.text_baseline = "top"

        if 'save_image' in kwargs:
            # define callback per
            # https://ipycanvas.readthedocs.io/en/latest/retrieve_images.html
            def save_to_file(*args, **kwargs):
                self.canvas.to_file(self.save_image)

            self.canvas.observe(save_to_file, "image_data")

    def draw_rectangle(self, coords, color):
        self.canvas.fill_style = color
        self.canvas.fill_rect(*coords)

    def draw_text(self, text, xpos, ypos, color="black", align="center"):
        self.canvas.fill_style = color
        self.canvas.text_align = align
        self.canvas.fill_text(text, xpos, ypos)

    def image(self):
        return self.canvas

    def save(self):
        # handled via callback, above.
        pass



"""encode observations

X is-transcribed in celltype/tissue/compartment M at time T
Z is-present in M at time T
X is-up in M at time T
Y is-down in M at time T
"""

_obs = []
def _add_obs(ob):
    assert isinstance(ob, Observation), f"{ob} must be an Observation"
    global _obs
    _obs.append(ob)

def get_obs():
    return list(_obs)

def reset():
    global _obs
    _obs = []

class Observation:
    pass

class Obs_IsPresent(Observation):
    def __init__(self, *, gene=None, time=None, tissue=None):
        assert gene, "gene must be set"
        assert time is not None, "time must be set"
        assert tissue is not None, "tissue must be set"
        self.gene_name = gene
        self.time = time
        self.tissue_name = tissue

    def check(self, state):
        # not applicable
        if state.time != self.time:
            return None

        tissue_state = state.get_by_tissue_name(self.tissue_name)
        return tissue_state.is_active(gene_name=self.gene_name)

    def render(self):
        return f"{self.gene_name} is ON in tissue {self.tissue_name} at time {self.time}"

def check_is_present(*, gene=None, time=None, tissue=None):
    ob = Obs_IsPresent(gene=gene, time=time, tissue=tissue)
    _add_obs(ob)



class Obs_IsNotPresent(Observation):
    def __init__(self, *, gene=None, time=None, tissue=None):
        assert gene
        assert time is not None
        assert tissue is not None
        self.gene_name = gene
        self.time = time
        self.tissue_name = tissue

    def check(self, state):
        # not applicable
        if state.time != self.time:
            return None

        tissue_state = state.get_by_tissue_name(self.tissue_name)
        return not tissue_state.is_active(gene_name=self.gene_name)

    def render(self):
        return f"{self.gene_name} is NOT ON in tissue {self.tissue_name} at time {self.time}"

def check_is_not_present(*, gene=None, time=None, tissue=None):
    ob = Obs_IsNotPresent(gene=gene, time=time, tissue=tissue)
    _add_obs(ob)



class Obs_IsNeverPresent(Observation):
    def __init__(self, *, gene=None, tissue=None):
        assert gene
        assert tissue is not None
        self.gene_name = gene
        self.tissue_name = tissue

    def check(self, state):
        tissue_state = state.get_by_tissue_name(self.tissue_name)
        return not tissue_state.is_active(gene_name=self.gene_name)

    def render(self):
        return f"{self.gene_name} is NEVER ON in tissue {self.tissue_name}"

def check_is_never_present(*, gene=None, tissue=None):
    ob = Obs_IsNeverPresent(gene=gene, tissue=tissue)
    _add_obs(ob)


class Obs_IsAlwaysPresent(Observation):
    def __init__(self, *, gene=None, tissue=None):
        assert gene
        assert tissue is not None
        self.gene_name = gene
        self.tissue_name = tissue

    def check(self, state):
        tissue_state = state.get_by_tissue_name(self.tissue_name)
        return tissue_state.is_active(gene_name=self.gene_name)

    def render(self):
        return f"{self.gene_name} is ALWAYS ON in tissue {self.tissue_name}"

def check_is_always_present(*, gene=None, tissue=None):
    ob = Obs_IsAlwaysPresent(gene=gene, tissue=tissue)
    _add_obs(ob)


def test_observations(state):
    succeed = True
    for ob in get_obs():
        check = ob.check(state)
        if check is None:
            pass
        elif check:
            print('passed:', ob.render())
        else:
            print('** FAILED:', ob.render())
            succeed = False

    return succeed



import pytest

import dinkum
from dinkum.exceptions import *
from dinkum.vfg import Gene
from dinkum.vfn import Tissue
from dinkum import Timecourse
from dinkum import observations


def test_delete_gene():
    # check that 'reset' clears the gene.
    dinkum.reset()

    x = Gene(name='X')

    dinkum.reset()

    m = Tissue(name='M')
    y = Gene(name='Y')

    # these should both fail
    with pytest.raises(DinkumInvalidGene):
        x.is_present(where=m, start=1)

    with pytest.raises(DinkumInvalidGene):
        y.activated_by(source=x)


def test_delete_tissue():
    # check that 'reset' clears the tissue
    dinkum.reset()

    m = Tissue(name='M')

    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')

    with pytest.raises(DinkumInvalidTissue):
        x.is_present(where=m, start=1)



__all__ = ["MultiTissuePanel",
           "TissueActivityPanel" ]

from .draw_ipycanvas import IpycanvasDrawer
from .draw_pillow import PillowDrawer
from dinkum import vfg

class MultiTissuePanel:
    """
    Draw multiple panels, representing multiple tissues & gene expression
    in each one.
    """
    def __init__(self, *, states=None, tissue_names=None, save_image=None,
                 canvas_type='pillow', genes_by_name=None):
        # create individual panels for each tissue
        self.panels = [ TissueActivityPanel(states=states, tissue_name=t,
                                            genes_by_name=genes_by_name)
                        for t in tissue_names ]

        self.save_image = save_image
        self.canvas_type = canvas_type

    def draw(self, gene_state_fn, level=100):
        """
        Draw the basic background canvas, upon which gene activation info
        will be displayed.
        """

        # determine overall panel size
        total_width = 0
        x_offsets = [0]
        max_height = 0

        for p in self.panels:
            width, height = p.estimate_panel_size()
            max_height = max(height, max_height)
            total_width += width
            x_offsets.append(total_width)

        # build the appropriate canvas type
        if self.canvas_type == 'ipycanvas':
            canvas = IpycanvasDrawer(width=total_width,
                                     height=max_height,
                                     save_image=self.save_image)
        else:
            assert self.canvas_type == 'pillow'
            canvas = PillowDrawer(width=total_width,
                                  height=max_height,
                                  save_image=self.save_image)

        # draw each tissue, with each collection of genes,
        # spread out horizontally
        for p, x_offset in zip(self.panels, x_offsets):
            # draw background
            d = p.draw_tissue(canvas, x_offset=x_offset)

            # draw time point/tissue/state
            gene_names = p.gene_names
            times = p.times
            d.draw(canvas, times, gene_names, gene_state_fn)

        if self.save_image:
            canvas.save()
            
        return canvas.image()


class Tissue_TimePointGene_Location:
    """
    A class to track each time point & gene location within a particular
    tissue.
    """
    def __init__(self, timepoint, gene, polygon_coords):
        self.tp = timepoint
        self.gene = gene
        self.polygon_coords = polygon_coords

    def draw(self, canvas, color):
        canvas.polygon(self.polygon_coords, fill=color)


class TissueActivityPanel:
    """
    A class to draw and display gene activity in a single tissue.

    Constructed to be used within a MultiTissuePanel, among other things.
    """
    box_size = 25
    box_spacing = 5
    
    box_x_start = 100
    box_y_start = 50
    
    def __init__(self, *, states=None, tissue_name=None, genes_by_name=None):
        assert tissue_name is not None
        self.tissue_name = tissue_name

        assert states is not None

        # determine all genes relevant to this tissue, + times, from 'states'.
        times = []
        all_gene_names = set()
        for (tp, state) in states:
            times.append(tp)
            
            activity = state.get_by_tissue_name(tissue_name)
            all_gene_names.update(activity.genes_by_name)

        ordered_names = []
        if not genes_by_name:
            ordered_names = list(all_gene_names)
        else:
            # prioritize genes_by_name; do remainder alphabetically
            for k in genes_by_name:
                ordered_names.append(k)
            for k in sorted(all_gene_names):
                if k not in genes_by_name:
                    ordered_names.append(k)
        self.gene_names = ordered_names

        self.times = times
        self.states = states

    def estimate_panel_size(self):
        "Estimate the size of this panel, based on # times / # genes"
        height = (len(self.times) + 1) * (self.box_size + self.box_spacing) + \
            self.box_y_start
        width = len(self.gene_names) * (self.box_size + self.box_spacing) + \
            self.box_x_start
        return width, height
    
    def draw_tissue(self, canvas, *, x_offset=0):
        """Draw this tissue on existing canvas.

        Returns a TissueActivityPanel_Draw that can be used to fill in the
        actual gene/time point/tissue activity.
        """
        gene_names = self.gene_names
        times = self.times

        box_total_size = self.box_size + self.box_spacing

        locations_by_tg = {}

        # determine the locations of the time points / genes.
        for row in range(0, len(times)):
            for col in range(0, len(gene_names)):
                xpos = self.box_x_start + box_total_size*col + x_offset
                ypos = self.box_y_start + box_total_size*row
                
                coords = [(xpos, ypos),
                          (xpos + self.box_size, ypos),
                          (xpos + self.box_size, ypos + self.box_size),
                          (xpos, ypos + self.box_size)]

                timep = times[row]
                gene_name = gene_names[col]
                loc = Tissue_TimePointGene_Location(timep, gene_name, coords)

                # save!
                locations_by_tg[(timep, gene_name)] = loc

        tissue_label_ypos = len(times)  * box_total_size + \
            self.box_y_start + 1/2*box_total_size
        tissue_label_xpos = self.box_x_start + x_offset + \
            round((box_total_size * len(gene_names)) / 2.0)

        canvas.draw_text(self.tissue_name, tissue_label_xpos,
                         tissue_label_ypos, align="center")
        self.locations_by_tg = locations_by_tg

        # draw row names / time points
        for row in range(0, len(times)):
            xpos = self.box_x_start - box_total_size / 2 + x_offset
            ypos = self.box_y_start + box_total_size*row
            canvas.draw_text(times[row], xpos, ypos, align="right")

        # draw col names / genes
        for col in range(0, len(gene_names)):
            ypos = self.box_y_start - box_total_size

            xpos = self.box_x_start + box_total_size*col
            xpos += x_offset

            canvas.draw_text(gene_names[col], xpos, ypos,
                             align="center")
                             #max_width = box_total_size)

        return TissueActivityPanel_Draw(self)


class TissueActivityPanel_Draw:
    "Use the timep/gene location to draw gene activity."
    active_color = "DeepSkyBlue" # blue!
    active_receptor_color = "DarkOliveGreen" # blue!

    present_color = (255, 0, 0)  # red!
    present_mask = (0, 255, 255) # ...?
    inactive_color = "DarkGrey"  # grey...
    
    def __init__(self, template):
        self.template = template

    def draw(self, canvas, times, gene_names, get_gene_state):
        locations_by_tg = self.template.locations_by_tg
        tissue_name = self.template.tissue_name

        for tp in times:
            for gene in gene_names:
                color = None

                active_color = self.active_color
                gs = get_gene_state(tissue_name, tp, gene)
                gene_obj = vfg.get_gene(gene)
                if gene_obj.is_receptor:
                    active_color = self.active_receptor_color

                if gs.active:
                    color = active_color
                elif gs.level > 0:
                    f = gs.level / 1000 # MAX_LEVEL @CTB
                    mask = self.present_mask
                    mask = tuple([ int(i * f) for i in mask ])
                    color = [ int(i+j) for (i, j) in zip(self.present_color, mask) ]
                    color = tuple(color)
                else:
                    color = self.inactive_color

                loc = locations_by_tg.get((tp, gene))
                if loc:
                    loc.draw(canvas, color)


###

class SeaUrchin_Blastula_ActivityPanel:
    def __init__(self, *, states=None, tissue_name=None):
        self.tissue_name = name
        self.states = states

    def draw_tissue(self, canvas, *, x_offset=0):
        pass



import pytest

import dinkum
from dinkum.vfg import Gene
from dinkum.vfn import Tissue
from dinkum import Timecourse
from dinkum import observations


def test_maternal():
    dinkum.reset()

    x = Gene(name='X')
    m = Tissue(name='M')
    x.is_present(where=m, start=1, duration=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='X', time=2, tissue='M')

    # run time course
    tc = dinkum.run(1, 5)
    assert len(tc) == 5


def test_maternal_fail():
    dinkum.reset()

    x = Gene(name='X')
    m = Tissue(name='M')
    x.is_present(where=m, start=1, duration=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_present(gene='X', time=2, tissue='M')

    with pytest.raises(dinkum.DinkumObservationFailed):
        dinkum.run(1, 5, verbose=True)


def test_activation():
    dinkum.reset()

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')

    y.activated_by(source=x)

    m = Tissue(name='M')
    x.is_present(where=m, start=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')

    # run!
    dinkum.run(start=1, stop=5)


def test_activation_fail():
    dinkum.reset()

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')

    y.activated_by(source=x)

    m = Tissue(name='M')
    x.is_present(where=m, start=1)

    # run!
    tc = dinkum.run(start=1, stop=5)

    # set observations
    observations.check_is_not_present(gene='X', time=1, tissue='M')
    with pytest.raises(dinkum.DinkumObservationFailed):
        tc.check()
    observations.reset()

    observations.check_is_present(gene='Y', time=1, tissue='M')
    with pytest.raises(dinkum.DinkumObservationFailed):
        tc.check()
    observations.reset()

    observations.check_is_not_present(gene='X', time=2, tissue='M')
    with pytest.raises(dinkum.DinkumObservationFailed):
        tc.check()
    observations.reset()

    observations.check_is_not_present(gene='Y', time=2, tissue='M')
    with pytest.raises(dinkum.DinkumObservationFailed):
        tc.check()
    observations.reset()


def test_simple_repression():
    dinkum.reset()

    # establish preconditions
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')

    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')
    observations.check_is_present(gene='Z', time=2, tissue='M')

    observations.check_is_not_present(gene='Z', time=3, tissue='M')

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')
    z = Gene(name='Z')

    y.activated_by(source=x)

    z.and_not(activator=x, repressor=y)

    m = Tissue(name='M')
    x.is_present(where=m, start=1)

    # run!
    dinkum.run(1, 5)


def test_simple_multiple_tissues():
    dinkum.reset()

    ## tissue M
    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_never_present(gene='Y', tissue='M')

    observations.check_is_present(gene='X', time=3, tissue='M')
    observations.check_is_present(gene='Z', time=3, tissue='M')

    ## tissue N
    observations.check_is_present(gene='Y', time=2, tissue='N')
    observations.check_is_present(gene='X', time=2, tissue='N')
    observations.check_is_present(gene='X', time=3, tissue='N')
    observations.check_is_present(gene='Y', time=3, tissue='N')
    observations.check_is_never_present(gene='Z', tissue='N')


    # set it all up!

    ## VFG
    a = Gene(name='A')
    b = Gene(name='B')

    x = Gene(name='X')
    y = Gene(name='Y')
    z = Gene(name='Z')

    x.activated_by(source=a)
    y.activated_by(source=b)
    z.and_not(activator=x, repressor=y)

    ## VFN
    m = Tissue(name='M')
    a.is_present(where=m, start=1)

    n = Tissue(name='N')
    a.is_present(where=n, start=1)
    b.is_present(where=n, start=1)

    # run!
    dinkum.run(1, 5)


def test_simple_positive_feedback():
    dinkum.reset()

    # establish preconditions
    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=2, tissue='M')

    observations.check_is_present(gene='X', time=3, tissue='M')
    observations.check_is_present(gene='Y', time=3, tissue='M')

    observations.check_is_present(gene='X', time=4, tissue='M')
    observations.check_is_present(gene='Y', time=4, tissue='M')

    # set it all up!
    a = Gene(name='A')
    x = Gene(name='X')
    y = Gene(name='Y')

    y.activated_by(source=x)
    x.activated_by_or(sources=[a, y])

    m = Tissue(name='M')
    a.is_present(where=m, start=1, duration=2)

    # run!
    dinkum.run(1, 5)


def test_simple_positive_feedback_old_name_activated_or():
    dinkum.reset()

    # establish preconditions
    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=2, tissue='M')

    observations.check_is_present(gene='X', time=3, tissue='M')
    observations.check_is_present(gene='Y', time=3, tissue='M')

    observations.check_is_present(gene='X', time=4, tissue='M')
    observations.check_is_present(gene='Y', time=4, tissue='M')

    # set it all up!
    a = Gene(name='A')
    x = Gene(name='X')
    y = Gene(name='Y')

    y.activated_by(source=x)
    # CTB this is an old name for 'activated_by_or'. Check that it works.
    x.activated_or(sources=[a, y])

    m = Tissue(name='M')
    a.is_present(where=m, start=1, duration=2)

    # run!
    dinkum.run(1, 5)


def test_simple_coherent_feed_forward():
    dinkum.reset()

    # establish preconditions
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_not_present(gene='Z', time=1, tissue='M')

    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')
    observations.check_is_not_present(gene='Z', time=2, tissue='M')

    observations.check_is_present(gene='X', time=3, tissue='M')
    observations.check_is_present(gene='Y', time=3, tissue='M')
    observations.check_is_present(gene='Z', time=3, tissue='M')

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')
    z = Gene(name='Z')

    y.activated_by(source=x)
    z.activated_by_and(sources=[x, y])

    m = Tissue(name='M')
    x.is_present(where=m, start=1)

    # run!
    dinkum.run(1, 5)


def test_simple_incoherent_feed_forward():
    dinkum.reset()

    # establish preconditions
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_not_present(gene='Z', time=1, tissue='M')

    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')
    observations.check_is_present(gene='Z', time=2, tissue='M')

    observations.check_is_present(gene='X', time=3, tissue='M')
    observations.check_is_present(gene='Y', time=3, tissue='M')
    observations.check_is_not_present(gene='Z', time=3, tissue='M')

    observations.check_is_present(gene='X', time=4, tissue='M')
    observations.check_is_present(gene='Y', time=4, tissue='M')
    observations.check_is_not_present(gene='Z', time=4, tissue='M')

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')
    z = Gene(name='Z')

    y.activated_by(source=x)
    z.and_not(activator=x, repressor=y)

    m = Tissue(name='M')
    x.is_present(where=m, start=1)

    # run!
    dinkum.run(1, 5)


def test_simple_incoherent_feed_forward_2_tissues():
    dinkum.reset()

    # establish preconditions
    observations.check_is_not_present(gene='Z', time=1, tissue='M')
    observations.check_is_present(gene='Z', time=2, tissue='M')
    observations.check_is_present(gene='Z', time=3, tissue='M')
    observations.check_is_present(gene='Z', time=4, tissue='M')

    observations.check_is_never_present(gene='Y', tissue='M')

    observations.check_is_not_present(gene='Z', time=1, tissue='N')
    observations.check_is_present(gene='Z', time=2, tissue='N')
    observations.check_is_not_present(gene='Z', time=3, tissue='N')
    observations.check_is_not_present(gene='Z', time=4, tissue='N')

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')
    z = Gene(name='Z')
    s = Gene(name='S')          # switches spec states b/t tissues M and N

    y.activated_by_and(sources=[x, s])
    z.and_not(activator=x, repressor=y)

    m = Tissue(name='M')
    x.is_present(where=m, start=1)

    m = Tissue(name='N')
    x.is_present(where=m, start=1)
    s.is_present(where=m, start=1)

    # run!
    dinkum.run(1, 5)


def test_simple_mutual_repression():
    dinkum.reset()

    # establish preconditions
    observations.check_is_never_present(gene='X', tissue='N')
    observations.check_is_never_present(gene='Y', tissue='M')

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')
    a = Gene(name='A')
    b = Gene(name='B')

    x.and_not(activator=a, repressor=y)
    y.and_not(activator=b, repressor=x)

    m = Tissue(name='M')
    a.is_present(where=m, start=1)

    n = Tissue(name='N')
    b.is_present(where=n, start=1)

    # run!
    dinkum.run(1, 5)


def test_simple_toggle_switch():
    dinkum.reset()

    # establish preconditions
    observations.check_is_not_present(gene='X', tissue='M', time=1)
    observations.check_is_present(gene='X', tissue='M', time=2)

    observations.check_is_never_present(gene='X', tissue='N')

    # set it all up!
    x = Gene(name='X')          # target of toggle switch
    t = Gene(name='T')          # toggle switch t.f.
    cofactor = Gene(name='cofactor')    # janus factor

    x.toggle_repressed(tf=t, cofactor=cofactor)

    # tissue M: t is present, cofactor is not; activate.
    m = Tissue(name='M')
    t.is_present(where=m, start=1)

    # tissue N: t is present, along with cofactor; repress.
    n = Tissue(name='N')
    t.is_present(where=n, start=1)
    cofactor.is_present(where=n, start=1)

    # run!
    dinkum.run(1, 5)


def test_delayed_activation():
    dinkum.reset()

    # set it all up!
    x = Gene(name='X')
    y = Gene(name='Y')

    y.activated_by(source=x, delay=2)

    m = Tissue(name='M')
    x.is_present(where=m, start=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=2, tissue='M')
    observations.check_is_present(gene='X', time=3, tissue='M')
    observations.check_is_present(gene='Y', time=3, tissue='M')

    # run!
    dinkum.run(start=1, stop=5)




import pytest

import dinkum
from dinkum.vfg import Gene, CustomActivation
from dinkum.vfn import Tissue
from dinkum import Timecourse
from dinkum import observations

from dinkum.exceptions import *
from dinkum.log import *


def test_custom_fn_1():
    # basic does-it-work
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    def activator_fn(*, X):
        if X.level == 0 or not X.active:
            assert bool(X) == False
        return X

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=activator_fn, delay=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')

    observations.check_is_not_present(gene='X', time=3, tissue='M')
    observations.check_is_not_present(gene='Y', time=3, tissue='M')

    # run time course
    tc = dinkum.run(1, 5)
    assert len(tc) == 5


def test_custom_fn_2():
    set_debug(True)

    # basic does-it-work
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    z = Gene(name='Z')
    m = Tissue(name='M')

    def activator_fn(*, Z, X):     # allow order independence
        return X

    x.is_present(where=m, start=1, duration=1)
    z.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=activator_fn, delay=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')

    observations.check_is_not_present(gene='X', time=3, tissue='M')
    observations.check_is_not_present(gene='Y', time=3, tissue='M')

    # run time course
    tc = dinkum.run(1, 5)
    assert len(tc) == 5


def test_custom_1_fail():
    # no gene Z
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    def activator_fn(*, Z):
        return X

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=activator_fn, delay=1)

    # run time course
    with pytest.raises(DinkumInvalidGene):
        tc = dinkum.run(1, 5)


def test_custom_bad_defn():
    # invalid activation function defns - must be kwargs
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    # must be explicitly named, not positional
    def activator_fn(Z):
        return X
    with pytest.raises(DinkumInvalidActivationFunction):
        y.custom_activation(state_fn=activator_fn, delay=1)

    # no defaults
    def activator_fn(Z=None):
        return X
    with pytest.raises(DinkumInvalidActivationFunction):
        y.custom_activation(state_fn=activator_fn, delay=1)

    # no defaults 2
    def activator_fn(*, Z=None):
        return X
    with pytest.raises(DinkumInvalidActivationFunction):
        y.custom_activation(state_fn=activator_fn, delay=1)

    # no general kwargs
    def activator_fn(**kwargs):
        return X
    with pytest.raises(DinkumInvalidActivationFunction):
        y.custom_activation(state_fn=activator_fn, delay=1)

    # no general args
    def activator_fn(*args):
        return X
    with pytest.raises(DinkumInvalidActivationFunction):
        y.custom_activation(state_fn=activator_fn, delay=1)


def test_custom_class_1():
    # does it work with a custom class? give a list of gene names
    set_debug(True)

    # basic does-it-work
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    class ActivateMe(CustomActivation):
        def __call__(self, *, X):
            return X

    state_fn = ActivateMe(input_genes=['X'])

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=state_fn, delay=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')

    observations.check_is_not_present(gene='X', time=3, tissue='M')
    observations.check_is_not_present(gene='Y', time=3, tissue='M')

    # run time course
    tc = dinkum.run(1, 5)
    assert len(tc) == 5


def test_custom_class_2():
    # does it work with a custom class? give a list of genes.
    set_debug(True)

    # basic does-it-work
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    class ActivateMe(CustomActivation):
        def __call__(self, *, X):
            return X

    state_fn = ActivateMe(input_genes=[x])

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=state_fn, delay=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')

    observations.check_is_not_present(gene='X', time=3, tissue='M')
    observations.check_is_not_present(gene='Y', time=3, tissue='M')

    # run time course
    tc = dinkum.run(1, 5)
    assert len(tc) == 5


def test_custom_class_3():
    # does it work with a custom class? retrieve gene names from __call__
    set_debug(True)

    # basic does-it-work
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    class ActivateMe(CustomActivation):
        def __call__(self, *, X):
            return X

    state_fn = ActivateMe()

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=state_fn, delay=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')

    observations.check_is_not_present(gene='X', time=3, tissue='M')
    observations.check_is_not_present(gene='Y', time=3, tissue='M')

    # run time course
    tc = dinkum.run(1, 5)
    assert len(tc) == 5


def test_custom_class_1_fail():
    # various failure modes
    set_debug(True)

    # basic does-it-work
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    class ActivateMe(CustomActivation):
        def __call__(self, *, Z):
            return X

    state_fn = ActivateMe()

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=state_fn, delay=1)

    # run time course
    with pytest.raises(DinkumInvalidGene):
        tc = dinkum.run(1, 5)



"""encode view-from-genome rules.
"""
from functools import total_ordering
import inspect
import collections

from .exceptions import *
from .vfn import check_is_valid_tissue

class GeneStateInfo:
    def __init__(self, level=0, active=False):
        self.level = level
        self.active = active

    def __bool__(self):
        return bool(self.level > 0 and self.active)

    def __iter__(self):
        return iter([self.level, self.active])

    def __repr__(self):
        return f"<level={self.level},active={self.active}>"

DEFAULT_OFF = GeneStateInfo(level=0, active=False)

_rules = []
_genes = []

def _add_rule(ix):
    _rules.append(ix)

    # check!
    seen = set()
    for r in _rules:
        if r.dest in seen and not r.multiple_allowed:
            raise DinkumException(f"multiple rules containing {r.dest} are not allowed!")
        if not r.multiple_allowed:
            seen.add(r.dest)


def get_rules():
    return list(_rules)

def get_gene_names():
    return [ g.name for g in sorted(_genes) ]

def get_gene(name):
    for g in sorted(_genes):
        if g.name == name:
            return g
    raise Exception(f"unknown genome name: '{name}'")

def reset():
    global _rules
    global _genes
    _rules = []
    _genes = []


def check_is_valid_gene(g):
    if not g in _genes:
        raise DinkumInvalidGene(f"{g.name} is invalid")

def check_is_tf(g):
    check_is_valid_gene(g)
    if not g.is_tf:
        raise DinkumNotATranscriptionFactor(f"{g.name} is not a transcription factor")


def _retrieve_ligands(timepoint, states, tissue, delay):
    "Retrieve all ligands in neighboring tissues for the given timepoint/delay"
    #assert isinstance(tissue, Tissue)

    ligands = set()
    for gene in _genes:
        if gene._is_ligand:
            for neighbor in tissue.neighbors:
                # for neighbor in (tissue, *tissue.neighbors):@CTB
                if states.is_active(timepoint, delay, gene, neighbor):
                    is_juxtacrine = getattr(gene, 'is_juxtacrine', False)
                    if is_juxtacrine:
                        if neighbor != tissue:
                            ligands.add(gene)
                    else:
                        ligands.add(gene)

    return ligands


class CustomActivation(object):
    def __init__(self, *, input_genes=None):
        # only support explicit kwargs on __call__,
        # because otherwise we are a bit
        # fragile with respect to order of arguments.
        argspec = inspect.getfullargspec(self.__call__)
        argspec.args.remove('self')
        if argspec.args or argspec.varargs or argspec.varkw or \
           argspec.kwonlydefaults:
            raise DinkumInvalidActivationFunction("on __call__, must supply _only_ kwargs with no defaults")

        if input_genes is None: # retrieve from __call__
            input_genes = argspec.kwonlyargs

        self.input_genes = list(input_genes)


class Interactions:
    multiple_allowed = False

    def btp_autonomus_links(self):
        raise Unimplemented

    def btp_signal_links(self):
        raise Unimplemented

    def check_ligand(self, timepoint, states, tissue, delay):
        if getattr(self.dest, '_set_ligand', None):
            #print(self.dest, "is a receptor w/ a ligand")
            ligands_in_neighbors = _retrieve_ligands(timepoint, states,
                                                     tissue, delay)
            if self.dest._set_ligand in ligands_in_neighbors:
                return True
            return False
        else:
            #print(self.dest, "is not a receptor")
            return True         # by default, not ligand => is active


class Interaction_IsPresent(Interactions):
    multiple_allowed = True

    def __init__(self, *, dest=None, start=None, duration=None, tissue=None,
                 level=None, decay=None):
        assert isinstance(dest, Gene), f"'{dest}' must be a Gene (but is not)"
        assert start is not None, "must provide start time"
        assert level is not None, "must provide level"
        assert decay is not None, "must provide decay"
        assert decay >= 1
        assert decay < 1e6
        assert tissue
        self.dest = dest
        self.tissue = tissue
        self.start = start
        self.duration = duration
        self.level = level
        self.decay = decay

    def btp_autonomous_links(self):
        return []

    def btp_signal_links(self):
        return []

    def advance(self, *, timepoint=None, states=None, tissue=None):
        # ignore states
        if tissue == self.tissue:
            if timepoint >= self.start:
                if self.duration is None or \
                   timepoint < self.start + self.duration: # active!
                    if self.check_ligand(timepoint, states, tissue, delay=1):
                        yield self.dest, GeneStateInfo(level=self.level,
                                                       active=True)
                    else:
                        yield self.dest, GeneStateInfo(level=self.level,
                                                       active=False)
        # we have no opinion on activity outside our tissue!

        self.level = round(self.level / self.decay + 0.5)


class Interaction_Activates(Interactions):
    def __init__(self, *, source=None, dest=None, delay=1):
        check_is_valid_gene(dest)
        check_is_tf(source)

        self.src = source
        self.dest = dest
        self.delay = delay

    def btp_autonomous_links(self):
        yield self.dest, self.src, "positive"

    def btp_signal_links(self):
        return []

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if its source was active 'delay' ticks ago.
        """
        if not states:
            return

        assert tissue
        assert timepoint is not None

        if states.is_active(timepoint, self.delay, self.src, tissue):
            is_active = self.check_ligand(timepoint, states, tissue, self.delay)
            yield self.dest, GeneStateInfo(level=100, active=is_active)


class Interaction_Or(Interactions):
    def __init__(self, *, sources=None, dest=None, delay=1):
        check_is_valid_gene(dest)
        for g in sources:
            check_is_tf(g)

        for tf in sources:
            if not tf.is_tf:
                raise DinkumNotATranscriptionFactor(tf.name)
        self.sources = sources
        self.dest = dest
        self.delay = delay

    def btp_autonomous_links(self):
        for src in self.sources:
            return [self.dest, self.src, "positive"]

    def btp_signal_links(self):
        return []

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if any of its sources were activate 'delay'
        ticks ago.
        """
        if not states:
            return

        assert tissue

        source_active = [ states.is_active(timepoint, self.delay, g, tissue)
                          for g in self.sources ]

        if any(source_active):
            is_active = self.check_ligand(timepoint, states, tissue, self.delay)
            yield self.dest, GeneStateInfo(level=100, active=is_active)


class Interaction_AndNot(Interactions):
    def __init__(self, *, source=None, repressor=None, dest=None, delay=1):
        check_is_valid_gene(dest)
        check_is_tf(source)
        check_is_tf(repressor)

        self.src = source
        self.repressor = repressor
        self.dest = dest
        self.delay = delay
        for tf in (self.src, self.repressor):
            if not tf.is_tf:
                raise DinkumNotATranscriptionFactor(tf.name)

    def btp_autonomous_links(self):
        yield [self.dest, self.src, "positive"]
        yield [self.dest, self.repressor, "negative"]

    def btp_signal_links(self):
        return []

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if its activator was active 'delay' ticks ago,
        and its repressor was _not_ active then.
        """
        if not states:
            return

        assert tissue

        src_is_active = states.is_active(timepoint, self.delay,
                                         self.src, tissue)
        repressor_is_active = states.is_active(timepoint, self.delay,
                                              self.repressor, tissue)

        if src_is_active and not repressor_is_active:
            is_active = self.check_ligand(timepoint, states, tissue, self.delay)
            yield self.dest, GeneStateInfo(level=100, active=is_active)


class Interaction_And(Interactions):
    def __init__(self, *, sources=None, dest=None, delay=1):
        check_is_valid_gene(dest)
        for g in sources:
            check_is_tf(g)

        self.sources = sources
        self.dest = dest
        self.delay = delay
        for tf in self.sources:
            if not tf.is_tf:
                raise DinkumNotATranscriptionFactor(tf.name)

    def btp_autonomous_links(self):
        for src in self.sources:
            yield [self.dest, src, "positive"]

    def btp_signal_links(self):
        return []

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if all of its sources were active 'delay' ticks
        ago.
        """
        if not states:
            return

        assert tissue

        source_active = [ states.is_active(timepoint, self.delay, g, tissue)
                          for g in self.sources ]

        if all(source_active):
            is_active = self.check_ligand(timepoint, states, tissue, self.delay)
            yield self.dest, GeneStateInfo(level=100, active=is_active)


class Interaction_ToggleRepressed(Interactions):
    def __init__(self, *, tf=None, cofactor=None, dest=None, delay=1):
        check_is_valid_gene(dest)
        check_is_tf(tf)
        check_is_valid_gene(cofactor)

        self.tf = tf
        if not self.tf.is_tf:
            raise DinkumNotATranscriptionFactor(self.tf.name)
        self.cofactor = cofactor
        self.dest = dest
        self.delay = delay

    def btp_signal_links(self):
        return []

    def btp_autonomous_links(self):
        yield [self.dest, self.tf, "positive"]
        yield [self.dest, self.cofactor, "positive"] # @CTB toggle

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if the tf was active and the cofactor was active
        'delay' ticks ago.
        """
        if not states:
            return

        assert tissue

        tf_active = states.is_active(timepoint, self.delay,
                                     self.tf, tissue)
        cofactor_active = states.is_active(timepoint, self.delay,
                                           self.cofactor, tissue)


        # @CTB refigure for receptor/ligand, yah?
        if tf_active and not cofactor_active:
            is_active = self.check_ligand(timepoint, states, tissue, self.delay)
            yield self.dest, GeneStateInfo(level=100, active=is_active)


class Interaction_Arbitrary(Interactions):
    """
    An interaction that supports arbitrary logic, + levels.
    """
    def __init__(self, *, dest=None, state_fn=None, delay=1):
        assert dest
        assert state_fn
        _check_gene_names = self._get_gene_names(state_fn)

        self.dest = dest
        self.state_fn = state_fn
        self.delay = delay


    def btp_autonomous_links(self):
        return []

    def btp_signal_links(self):
        return []

    def _get_gene_names(self, state_fn):
        # get the names of the genes on the function
        if isinstance(state_fn, CustomActivation):
            dep_gene_names = state_fn.input_genes

            # allow genes, or gene names
            dep_gene_names = [ g.name if isinstance(g, Gene) else g
                               for g in dep_gene_names ]
        else:
            # only support explicit kwargs, because otherwise we are a bit
            # fragile with respect to order of arguments.
            argspec = inspect.getfullargspec(state_fn)
            if argspec.args or argspec.varargs or argspec.varkw or \
               argspec.kwonlydefaults:
                raise DinkumInvalidActivationFunction("must supply kwargs only")
            dep_gene_names = argspec.kwonlyargs

        return dep_gene_names

    def _get_genes_for_activation_fn(self, state_fn):
        # get the activity of upstream genes
        dep_genes = []
        dep_gene_names = self._get_gene_names(state_fn)
        for name in dep_gene_names:
            found = False
            for g in _genes:
                if g.name == name:
                    check_is_tf(g)
                    dep_genes.append((name, g))
                    found = True
                    break
            if not found:
                raise DinkumInvalidGene(name)

        return dep_genes

    def advance(self, *, timepoint=None, states=None, tissue=None):
        # 'states' is class States...
        if not states:
            return

        assert tissue
        dep_genes = self._get_genes_for_activation_fn(self.state_fn)

        # pass in their full GeneStateInfo
        delay = self.delay
        dep_state = { name: states.get_gene_state_info(timepoint, delay, g, tissue)
                      for (name, g) in dep_genes }

        result = self.state_fn(**dep_state)
        if result is not None:
            if not isinstance(result, GeneStateInfo):
                try:
                    if len(tuple(result)) == 2:
                        result = GeneStateInfo(int(result[0]), bool(result[1]))
                except:
                    pass
            if not isinstance(result, GeneStateInfo):
                raise DinkumInvalidActivationResult(f"result '{result}' of custom activation function is not a GeneStateInfo tuple (and cannot be converted)")

        if result is not None:
            level, is_active = result

            if is_active:
                # @CTB ...check ligand should be made more general...
                is_active = self.check_ligand(timepoint,
                                              states,
                                              tissue,
                                              self.delay)

            yield self.dest, GeneStateInfo(level, is_active)


class Gene:
    is_receptor = False
    is_tf = True

    def __init__(self, *, name=None):
        global _genes

        assert name, "Gene must have a name"
        self.name = name

        _genes.append(self)
        self._set_ligand = None
        self._is_ligand = None

    def __repr__(self):
        return f"Gene('{self.name}')"

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return self.name != other.name

    def __lt__(self, other):
        return self.name < other.name

    def __hash__(self):
        return hash(self.name)

    def present(self):
        return 1

    def ligand_present(self):
        return (self._set_ligand is None) or 1 # @CTB

    def active(self):           # present = active
        if self._is_ligand:
            return self.present() and self.ligand_present()
        else:
            return self.present()

    def activated_by(self, *, source=None, delay=1):
        check_is_valid_gene(self)
        check_is_valid_gene(source)
        ix = Interaction_Activates(source=source, dest=self, delay=delay)
        _add_rule(ix)

    def activated_by_or(self, *, sources=None, delay=1):
        ix = Interaction_Or(sources=sources, dest=self, delay=delay)
        _add_rule(ix)
    activated_or = activated_by_or

    def and_not(self, *, activator=None, repressor=None, delay=1):
        ix = Interaction_AndNot(source=activator, repressor=repressor,
                                dest=self, delay=delay)
        _add_rule(ix)

    def activated_by_and(self, *, sources, delay=1):
        ix = Interaction_And(sources=sources, dest=self, delay=delay)
        _add_rule(ix)

    def toggle_repressed(self, *, tf=None, cofactor=None, delay=1):
        ix = Interaction_ToggleRepressed(tf=tf, cofactor=cofactor,
                                         dest=self, delay=delay)
        _add_rule(ix)

    def is_present(self, *, where=None, start=None, duration=None, level=100,
                   decay=1):
        assert where
        assert start
        check_is_valid_gene(self)
        check_is_valid_tissue(where)
        ix = Interaction_IsPresent(dest=self, start=start, duration=duration,
                                   tissue=where, level=level, decay=decay)
        _add_rule(ix)

    def custom_activation(self, *, state_fn=None, delay=1):
        ix = Interaction_Arbitrary(dest=self, state_fn=state_fn, delay=delay)
        _add_rule(ix)


class Ligand(Gene):
    is_tf = False

    def __init__(self, *, name=None, is_juxtacrine=False):
        super().__init__(name=name)
        self.is_juxtacrine = is_juxtacrine
        self._is_ligand = True

    def __repr__(self):
        return f"Ligand('{self.name}')"


class Receptor(Gene):
    is_receptor = True

    def __init__(self, *, name=None, ligand=None):
        super().__init__(name=name)
        assert name
        self._set_ligand = ligand
        if ligand:
            ligand._is_ligand = True

    def __repr__(self):
        return f"Receptor('{self.name}')"

    def ligand(self, *, activator=None, ligand=None):
        # assert 0, "legacy method!"
        # @CTB legacy
        if ligand is None:
            ligand = self._set_ligand
            if ligand is None:
                raise Exception("need to specify a ligand for this receptor, either at creation or here")
        else:
            ligand._is_ligand = True

        ix = Interaction_Activates(source=activator, dest=self, delay=1)
        _add_rule(ix)

    def activated_by(self, *, activator=None, source=None, delay=1):
        if activator is None:   # @CTB deprecated
            # assert 0
            activator = source
        if activator is None:
            raise Exception("must supply an activator!")

        ix = Interaction_Activates(source=source, dest=self, delay=delay)
        _add_rule(ix)



import csv

from . import vfg, vfn


def output_biotapestry_csv(output_filename):
    def fill_ten(ls):
        return list(ls) + ['']*(10 - len(ls))

    fp = open(output_filename, 'w', newline='')
    w = csv.writer(fp)

    w.writerow(fill_ten(['# Model Commands']))
    w.writerow(fill_ten(['# Command Type', 'Model Name', 'Parent Model']))
    w.writerow(fill_ten(["model", "root"]))

    base_name = "double neg"
    w.writerow(fill_ten(["model", base_name, "root"]))

    for tissue in vfn._tissues:
        w.writerow(fill_ten(["model", tissue.name, base_name]))

    w.writerow(fill_ten(['# Region Commands']))
    w.writerow(fill_ten(['# Command Type', 'Model Name', 'Region Name', 'Region Abbreviation']))

    for n, tissue in enumerate(vfn._tissues):
        w.writerow(fill_ten(["region", base_name, f"a{n}", f"a{n}"]))

    for n, tissue in enumerate(vfn._tissues):
        w.writerow(fill_ten(["region", tissue.name, f"a{n}", f"a{n}"]))

    w.writerow(fill_ten(['# Standard Interactions']))
    w.writerow(fill_ten(["# Command Type","Model Name","Source Type","Source Name","Target Type","Target Name","Sign","Source Region Abbrev","Target Region Abbrev"]))

    for ix in vfg._rules:
        model_name = base_name
        source_type = 'gene'
        source_region = 'a0'
        target_region = 'a0'
        for (target, source, sign) in ix.btp_autonomous_links():
            source_name = source.name
            target_name = target.name
            assert sign in ["positive", "negative"]

            w.writerow(["general",model_name, source_type, source_name, 'gene',
                        target_name, sign, source_region, target_region])

    w.writerow(fill_ten(["# Signals"]))
    w.writerow(fill_ten(["# Standalone nodes"]))
    w.writerow(fill_ten(["# Interactions for Submodels"]))
    w.writerow(["# Command Type","Model Name","Source Type","Source Name","Target Type","Target Name","Sign","Source Region Abbrev","Target Region Abbrev",])
    fp.close()



"""
Top-level dinkum module
"""
import sys
from importlib.metadata import version
import pandas as pd

__version__ = version("dinkum-bio")

import itertools
import collections

from . import vfg
from .vfg import GeneStateInfo, DEFAULT_OFF
from . import vfn
from . import observations
from . import utils
from .exceptions import *

def reset(*, verbose=True):
    vfg.reset()
    vfn.reset()
    observations.reset()
    if verbose:
        print(f"initializing: dinkum v{__version__}")


def run_and_display_df(*, start=1, stop=10, gene_names=None, tissue_names=None,
                       verbose=False, save_image=None):
    """
    Run and display the circuit model.

    Key parameter:
    - start (default: 1)
    - stop (default 10)

    Other parameters:
    - 'tissue_names' - a list of tissue names to display (default: all).
    - 'gene_names' - a list of gene names to display (default: all).
    - 'verbose' - display more text output.
    - 'save_image' - save image to this file.
    - 'canvas_type' - 'ipycanvas' or 'pillow' (default: 'pillow')
    """
    from dinkum.display import MultiTissuePanel, tc_record_activity
    if not gene_names:
        gene_names = vfg.get_gene_names()
    if not tissue_names:
        tissue_names = vfn.get_tissue_names()

    try:
        states, tissues, get_state_fn = \
            tc_record_activity(start=start,
                               stop=stop,
                               gene_names=gene_names,
                               verbose=verbose)
    except DinkumException as e:
        print(f"ERROR: {str(e)}", file=sys.stderr)
        print("Halting execution.", file=sys.stderr)
        return None, None, None

    level_df, active_df = convert_states_to_dataframe(states, gene_names,
                                                      get_state_fn)

    mp = MultiTissuePanel(states=states, tissue_names=tissue_names,
                          genes_by_name=gene_names,
                          save_image=save_image)
    return mp.draw(get_state_fn), level_df, active_df


def run_and_display(*args, **kwargs):
    display_obj, _level_df, _active_df = run_and_display_df(*args, **kwargs)
    return display_obj


class GeneStates:
    def __init__(self):
        self.genes_by_name = {}

    def __repr__(self):
        return repr(self.genes_by_name)

    def __getitem__(self, gene_name):
        return self.get_gene_state(gene_name)

    def set_gene_state(self, *, gene=None, state_info=None):
        assert gene is not None
        assert state_info is not None
        self.genes_by_name[gene.name] = state_info

    def is_active(self, gene_name):
        state_info = self.genes_by_name.get(gene_name, DEFAULT_OFF)
        return state_info.active

    def get_level(self, gene_name):
        state_info = self.genes_by_name.get(gene_name, DEFAULT_OFF)
        return state_info.level

    def get_gene_state(self, gene_name):
        state_info = self.genes_by_name.get(gene_name, DEFAULT_OFF)
        return state_info

    def __contains__(self, gene):
        return self.is_active(gene.name)

    def report_activity(self):
        rl = []
        for k, v in sorted(self.genes_by_name.items()):
            if v:
                v = 1
            else:
                v = 0
            rl.append(f"{k}={v}")

        return rl


class TissueAndGeneStateAtTime:
    """
    Hold the gene activity state for multiple tissues at a particular tp.

    Holds multiple tissue, each with their own GeneStates object.

    dict interface supports getting and setting gene activity (value) by
    tissue (key).

    `tissues` attribute provides enumeration of available tissues

    `is_active(gene, tissue)` returns True/False around activity.
    """
    def __init__(self, *, tissues=None, time=None):
        assert tissues is not None
        assert time is not None

        self._tissues = list(tissues)
        self._tissues_by_name = {}
        self.time = time

    def __setitem__(self, tissue, genes):
        "set Tissue object by name."
        assert tissue in self._tissues
        assert isinstance(genes, GeneStates)
        self._tissues_by_name[tissue.name] = genes

    def __getitem__(self, tissue):
        "get Tissue object."
        return self._tissues_by_name[tissue.name]

    def get_by_tissue_name(self, tissue_name):
        return self._tissues_by_name[tissue_name]

    @property
    def tissues(self):
        return self._tissues

    def is_active(self, gene, tissue):
        ts = self[tissue]
        if ts and gene in ts:
            return True
        return False

    def get_gene_state_info(self, gene, tissue):
        ts = self[tissue]
        return ts[gene.name]


class TissueGeneStates(collections.UserDict):
    """
    Contains (potentially incomplete) set of tissue/gene states for many
    timepoints.
    """
    def __init__(self):
        self.data = {}

    def is_active(self, current_tp, delay, gene, tissue):
        from .vfg import Gene

        assert int(current_tp)
        assert int(delay)
        assert isinstance(gene, Gene)

        check_tp = current_tp - delay
        state = self.get(check_tp)
        if state and state.is_active(gene, tissue):
            return True
        return False

    def get_gene_state_info(self, current_tp, delay, gene, tissue):
        from .vfg import Gene

        assert int(current_tp)
        assert int(delay)
        assert isinstance(gene, Gene)

        check_tp = current_tp - delay
        time_state = self.get(check_tp)
        if time_state:
            return time_state.get_gene_state_info(gene, tissue)
        return None


class Timecourse:
    """
    Run a time course for a system b/t two time points, start and stop.
    """
    def __init__(self, *, start=None, stop=None):
        assert start is not None
        assert stop is not None

        print(f'start={start} stop={stop}')

        self.start = start
        self.stop = stop
        self.states_d = TissueGeneStates()

    def __iter__(self):
        return iter(self.states_d.values())

    def __len__(self):
        return len(self.states_d)

    def run(self, *, verbose=False):
        start = self.start
        stop = self.stop

        tissues = vfn.get_tissues()
        if verbose:
            print(f"got {len(tissues)} tissues.")
            for t in tissues:
                print(f"\ttissue {t.name}")
            print('')

        # advance one tick at a time
        this_state = {}
        for tp in range(start, stop + 1):
            next_state = TissueAndGeneStateAtTime(tissues=tissues, time=tp)

            for t in tissues:
                seen = set()
                next_active = GeneStates()
                for r in vfg.get_rules():
                    # advance state of all genes based on last state
                    for g, state_info in r.advance(timepoint=tp,
                                                   states=self.states_d,
                                                   tissue=t):

                        next_active.set_gene_state(gene=g,
                                                   state_info=state_info)


                next_state[t] = next_active
                #print('fff', t, next_active)

            # advance => next state
            self.states_d[tp] = next_state
            this_state = next_state

    def check(self):
        "Test all of the observations for all of the states."
        for state in iter(self):
            if not observations.test_observations(state):
                raise DinkumObservationFailed(state.time)


def run(start, stop, *, verbose=False):
    # run time course
    tc = Timecourse(start=start, stop=stop)
    tc.run(verbose=verbose)

    for state in tc:
        print(f"time={state.time}")
        for ti in state.tissues:
            present = state[ti]
            print(f"\ttissue={ti.name}, {present.report_activity()}")
        if not observations.test_observations(state):
            raise DinkumObservationFailed(state.time)

    return tc


def convert_states_to_dataframe(states, gene_names, get_state_fn):
    level_rows = []
    active_rows = []

    for tp, tissue_and_gene_sat in states:
        assert tp.startswith('t=')
        timepoint = int(tp.split('=')[1]) # convert tp str to int

        # for each tissue, get level of each gene
        for tissue in tissue_and_gene_sat.tissues:
            level_d = dict(tissue=tissue.name, timepoint=timepoint, timepoint_str=tp)
            active_d = dict(tissue=tissue.name, timepoint=timepoint, timepoint_str=tp)
            for gene_name in gene_names:
                gsa = get_state_fn(tissue.name, tp, gene_name)
                level_d[gene_name] = gsa.level
                active_d[gene_name] = gsa.active
            level_rows.append(level_d)
            active_rows.append(active_d)

    level_df = pd.DataFrame.from_dict(level_rows)
    active_df = pd.DataFrame.from_dict(active_rows)

    return level_df, active_df



import dinkum
from dinkum.vfg import Gene, Receptor, Ligand
from dinkum.vfn import Tissue
from dinkum import Timecourse
from dinkum import vfg, vfn


def define_model():
    dinkum.reset()

    # gataE, gcm, foxA

    # set it all up!
    pmar1 = Gene(name='pmar1')
    hesC = Gene(name='hesC')
    alx1 = Gene(name='alx1')
    delta = Ligand(name='delta')
    notch = Receptor(name='su(h)', ligand=delta)
    tbr = Gene(name='tbr')
    ets1 = Gene(name='ets1')
    gataE = Gene(name='gataE')
    gcm = Gene(name='gcm')
    foxA = Gene(name='foxA')

    embryo = Tissue(name='rest of embryo')
    micromere = Tissue(name='micromeres')
    embryo.add_neighbor(neighbor=micromere)

    # maternal genes
    early_ubiq = Gene(name='ub1')
    late_ubiq = Gene(name='ub2')

    ## set up maternal gene expression

    # early ubiq
    early_ubiq.is_present(where=micromere, start=1)
    early_ubiq.is_present(where=embryo, start=1)

    # late ubiq
    late_ubiq.is_present(where=micromere, start=4)
    late_ubiq.is_present(where=embryo, start=4)

    # pmar1: maternal in micromeres only
    pmar1.is_present(where=micromere, start=1)

    # hesC: present in both at the beginning
    hesC.is_present(where=micromere, start=1, duration=1)
    hesC.is_present(where=embryo, start=1, duration=1)

    # notch: present everywhere (?)
    notch.is_present(where=micromere, start=1)
    notch.is_present(where=embryo, start=1)

    ## set up all downstream genes

    # hesC: early, if not for pmar1
    hesC.and_not(activator=early_ubiq, repressor=pmar1)

    # alx &c: late, if not for hesC
    alx1.and_not(activator=late_ubiq, repressor=hesC)
    delta.and_not(activator=late_ubiq, repressor=hesC)
    tbr.and_not(activator=late_ubiq, repressor=hesC)
    ets1.and_not(activator=late_ubiq, repressor=hesC)

    # gataE, gcm, foxA: all activated by notch
    gataE.activated_by_and(sources=[late_ubiq, notch])
    gcm.activated_by_and(sources=[late_ubiq, notch])
    foxA.activated_by_and(sources=[late_ubiq, notch])


def test_output_double_neg():
    filename = '/tmp/out.btp.csv'
    define_model()
    print(f'writing to {filename}')
    dinkum.utils.output_biotapestry_csv('/tmp/out.btp.csv')



class DinkumException(Exception):
    pass

class DinkumObservationFailed(DinkumException):
    pass

class DinkumInvalidGene(DinkumException):
    pass

class DinkumNotATranscriptionFactor(DinkumException):
    pass

class DinkumInvalidTissue(DinkumException):
    pass

class DinkumNoSuchGene(DinkumException):
    pass

class DinkumInvalidActivationFunction(DinkumException):
    pass

class DinkumInvalidActivationResult(DinkumException):
    pass
